'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.combineReducers = exports.demux = exports.node = undefined;

var _iterator5 = require('babel-runtime/core-js/symbol/iterator');

var _iterator6 = _interopRequireDefault(_iterator5);

var _typeof2 = require('babel-runtime/helpers/typeof');

var _typeof3 = _interopRequireDefault(_typeof2);

var _set = require('babel-runtime/core-js/set');

var _set2 = _interopRequireDefault(_set);

var _getIterator2 = require('babel-runtime/core-js/get-iterator');

var _getIterator3 = _interopRequireDefault(_getIterator2);

var _toConsumableArray2 = require('babel-runtime/helpers/toConsumableArray');

var _toConsumableArray3 = _interopRequireDefault(_toConsumableArray2);

var _map = require('babel-runtime/core-js/map');

var _map2 = _interopRequireDefault(_map);

var _defineProperty = require('babel-runtime/core-js/object/define-property');

var _defineProperty2 = _interopRequireDefault(_defineProperty);

var _symbol = require('babel-runtime/core-js/symbol');

var _symbol2 = _interopRequireDefault(_symbol);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

exports.default = combiner;

var _combiner = combiner(defaultCombineReducers),
    node = _combiner.node,
    demux = _combiner.demux;

exports.node = node;
exports.demux = demux;
var combineReducers = exports.combineReducers = defaultCombineReducers;

var ACTIONS = (0, _symbol2.default)('actions');
var REDUCERS = (0, _symbol2.default)('reducers');

function combiner(combineReducers) {

    return {
        node: initNode(combineReducers),
        demux: initDemux(combineReducers)
    };
}

function initNode(combineReducers) {

    return function nodeFn(initial) {

        checkInitial(initial);

        var innerReducer = getInnerReducer(initial, combineReducers);

        var fn = function fn(state, action) {

            state = applyOwnReducers(state, action, fn[REDUCERS]);
            state = innerReducer(state, action);

            return state;
        };

        defineActions(fn, innerReducer[ACTIONS]);
        (0, _defineProperty2.default)(fn, REDUCERS, { value: new _map2.default() });
        Object.defineProperty(fn, 'on', { value: addActionsReducers });

        return fn;
    };
}

function initDemux(combineReducers) {

    return function demuxFn(initial, schema, selector) {

        checkInitial(initial);

        var selectChild = createSelector(selector);
        var innerReducer = getInnerReducer(initial, combineReducers);
        var childReducer = getChildReducer(schema, selectChild, combineReducers);

        var fn = function fn(state, action) {

            state = applyOwnReducers(state, action, fn[REDUCERS]);
            state = innerReducer(state, action);
            state = childReducer(state, action);

            return state;
        };

        var fnActions = !childReducer[ACTIONS] || !innerReducer[ACTIONS] ? null : [].concat((0, _toConsumableArray3.default)(childReducer[ACTIONS]), (0, _toConsumableArray3.default)(innerReducer[ACTIONS]));

        defineActions(fn, fnActions);
        (0, _defineProperty2.default)(fn, REDUCERS, { value: new _map2.default() });
        Object.defineProperty(fn, 'on', { value: addActionsReducers });

        return fn;
    };
}

function checkInitial(initial) {
    if (typeof initial === 'undefined') {
        throw new Error('node must be initialized.');
    }
}

function applyOwnReducers(state, action, actionReducers) {

    var reducers = actionReducers.get(action.type);

    if (!reducers) {
        return state;
    }

    return reducers.reduce(function (state, reducer) {
        return reducer(state, action);
    }, state);
}

function getInnerReducer(initial, combineReducers, noWrap) {

    if (isFunction(initial)) {
        return initial;
    }

    if (!isObject(initial)) {
        return noWrap ? initial : makeActionIdentity(initial);
    }

    var _getInitialStateReduc = getInitialStateReducers(initial, combineReducers),
        reducers = _getInitialStateReduc.reducers,
        defaultReducers = _getInitialStateReduc.defaultReducers;

    if (reducers) {
        return getInitialInnerReducer(initial, reducers, defaultReducers, combineReducers);
    } else {
        return noWrap ? initial : makeActionIdentity(initial);
    }
}

function getInitialStateReducers(initial, combineReducers) {

    var reducers = {};
    var defaultReducers = isArray(initial) ? [] : {};
    var hasFn = false;

    for (var k in initial) {
        if (hasOwn(initial, k)) {
            (function () {

                var innerReducer = getInnerReducer(initial[k], combineReducers, true);

                if (isFunction(innerReducer)) {
                    hasFn = true;
                    defaultReducers[k] = reducers[k] = innerReducer;
                } else {
                    defaultReducers[k] = function () {
                        return innerReducer;
                    };
                }
            })();
        }
    }

    return { reducers: hasFn ? reducers : null, defaultReducers: defaultReducers };
}

function getInitialInnerReducer(initial, reducers, defaultReducers, combineReducers) {

    var getDefault = combineReducers(defaultReducers);

    var fn = function fn(state, action) {

        if (state === undefined) {
            return getDefault(undefined, action);
        }

        if (!hasAction(fn, action)) {
            return state;
        }

        var stateReducers = {};

        for (var k in state) {
            if (hasOwn(state, k)) {
                if (hasAction(reducers[k], action)) {
                    stateReducers[k] = reducers[k];
                } else {
                    stateReducers[k] = identity;
                }
            }
        }

        var reducer = combineReducers(stateReducers);

        return reducer(state, action);
    };

    var reducersActions = getReducersActions(reducers);
    defineActions(fn, reducersActions);

    return fn;
}

function getChildReducer(schema, selectChild, combineReducers) {

    var schemaReducer = getInnerReducer(schema, combineReducers);

    var fn = function fn(state, action) {

        if (!isObject(state) || !hasAction(fn, action)) {
            return state;
        }

        var reducers = getStateChildReducers(state, action, selectChild, schemaReducer);

        var reducer = reducers ? combineReducers(reducers) : identity;

        return reducer(state, action);
    };

    defineActions(fn, schemaReducer[ACTIONS]);

    return fn;
}

function getStateChildReducers(state, action, selectChild, schemaReducer) {

    var reducers = mapIdenity(state);

    var childKey = selectChild(state, action);

    if (isIterableObject(childKey)) {
        var someKeysAreSet = false;
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = (0, _getIterator3.default)(childKey), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var k = _step.value;

                if (hasOwn(state, k)) {
                    someKeysAreSet = true;
                    reducers[k] = schemaReducer;
                }
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        if (!someKeysAreSet) {
            return null;
        }
    } else {
        if (!hasOwn(state, childKey)) {
            return null;
        }
        reducers[childKey] = schemaReducer;
    }

    return reducers;
}

function getReducersActions(reducers) {

    var actions = [];

    for (var k in reducers) {
        var reducer = reducers[k];
        if (!reducer[ACTIONS]) {
            return null;
        } else {
            actions = [].concat((0, _toConsumableArray3.default)(actions), (0, _toConsumableArray3.default)(reducer[ACTIONS]));
        }
    }

    return actions;
}

function mapIdenity(obj) {

    var map = {};

    for (var k in obj) {
        if (hasOwn(obj, k)) {
            map[k] = identity;
        }
    }

    return map;
}

function createSelector() {
    var selector = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};


    if (isFunction(selector)) {
        return selector;
    }

    if (!isArray(selector)) {

        if (isObject(selector)) {
            return getDefaultSelector;
        }

        selector = isString(selector) ? selector.split('.') : [selector];
    }

    return function (state, action) {
        return getValueByPath(action, selector);
    };
}

function getDefaultSelector(state, action) {
    return isArray(state) ? action.index : action.id;
}

function getValueByPath(obj, path) {

    var pointer = obj;

    var _iteratorNormalCompletion2 = true;
    var _didIteratorError2 = false;
    var _iteratorError2 = undefined;

    try {
        for (var _iterator2 = (0, _getIterator3.default)(path), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
            var p = _step2.value;


            if (!hasOwn(pointer, p)) {
                return;
            }

            pointer = pointer[p];
        }
    } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
    } finally {
        try {
            if (!_iteratorNormalCompletion2 && _iterator2.return) {
                _iterator2.return();
            }
        } finally {
            if (_didIteratorError2) {
                throw _iteratorError2;
            }
        }
    }

    return pointer;
}

function addActionsReducers(actionTypes, reducers) {

    if (!isArray(actionTypes)) {
        actionTypes = [actionTypes];
    }

    reducers = getReducers(reducers);

    var _iteratorNormalCompletion3 = true;
    var _didIteratorError3 = false;
    var _iteratorError3 = undefined;

    try {
        for (var _iterator3 = (0, _getIterator3.default)(actionTypes), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
            var actionType = _step3.value;


            var actionReducers = this[REDUCERS].get(actionType);

            if (!actionReducers) {
                actionReducers = [];
                this[ACTIONS] && this[ACTIONS].add(actionType);
                this[REDUCERS].set(actionType, actionReducers);
            }

            var _iteratorNormalCompletion4 = true;
            var _didIteratorError4 = false;
            var _iteratorError4 = undefined;

            try {
                for (var _iterator4 = (0, _getIterator3.default)(reducers), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
                    var reducer = _step4.value;

                    actionReducers.push(reducer);
                }
            } catch (err) {
                _didIteratorError4 = true;
                _iteratorError4 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion4 && _iterator4.return) {
                        _iterator4.return();
                    }
                } finally {
                    if (_didIteratorError4) {
                        throw _iteratorError4;
                    }
                }
            }
        }
    } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
    } finally {
        try {
            if (!_iteratorNormalCompletion3 && _iterator3.return) {
                _iterator3.return();
            }
        } finally {
            if (_didIteratorError3) {
                throw _iteratorError3;
            }
        }
    }

    return this;
}

function getReducers(reducers) {

    if (reducers === undefined) {
        return [];
    }

    if (isFunction(reducers)) {
        return [reducers];
    }

    if (!isArray(reducers)) {
        return [function () {
            return reducers;
        }];
    }

    if (reducers.some(isFunction)) {
        return reducers.map(function (reducer) {
            return isFunction(reducer) ? reducer : function () {
                return reducer;
            };
        });
    }

    return [function () {
        return reducers;
    }];
}

function defaultCombineReducers(reducersMap) {

    return function () {
        var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
        var action = arguments[1];


        var result = isArray(state) ? [] : isArray(reducersMap) ? [] : {};

        var changed = false;

        for (var k in reducersMap) {

            var reducer = reducersMap[k];

            var prevState = state[k];
            var nextState = reducer(prevState, action);

            if (typeof nextState === 'undefined') {
                throw new Error('undefined state returned for key ' + ('"' + k + '" on "' + action.type + '" action.'));
            }

            result[k] = nextState;
            changed = changed || prevState !== nextState;
        }

        return changed ? result : state;
    };
}

function defineActions(fn, actions) {
    if (!actions) {
        return;
    }
    (0, _defineProperty2.default)(fn, ACTIONS, {
        value: new _set2.default(actions)
    });
}

function hasAction(fn, action) {
    return fn && (!fn[ACTIONS] || fn[ACTIONS].has(action.type));
}

function makeActionIdentity(initial) {
    var fn = function fn() {
        var data = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : initial;

        return data;
    };
    defineActions(fn, []);
    return fn;
}

function identity() {
    var data = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

    return data;
}

function isObject(obj) {
    return (typeof obj === 'undefined' ? 'undefined' : (0, _typeof3.default)(obj)) === 'object' && obj;
}

function isArray(arr) {
    return Array.isArray(arr);
}

function isFunction(fn) {
    return typeof fn === 'function';
}

function isString(str) {
    return typeof str === 'string';
}

function isIterableObject(iterable) {
    return isObject(iterable) && isFunction(iterable[_iterator6.default]);
}

function hasOwn(obj, prop) {
    return Object.prototype.hasOwnProperty.call(obj, prop);
}